#ifndef EXAMPLE_CONFIG_H
#define EXAMPLE_CONFIG_H
/*!
 * @file
 * @brief This is a config file
 * @details This is a detailed description
 * @ingroup animals
 * @addtogroup animals
 * @{
 */

#define CONFIG_HELLO (123)
#define CONFIG_WORLD ("abx")
#define PI 3.14159265358979323846
#define PRINT_PRETTY(MSG, ...) printf(MSG, __VA_ARGS__)

/*! @} */
#endif
