
/**
 * @brief print function
 *
 */
void print(void);

/**
 * @brief print the added result
 *
 * @param a a
 * @param b b
 */
void add(int a, int b);


/**
 * @brief struct comp2
 *
 */
struct comp2 {
    int foo;
    int bar;
};


/**
 * @brief typedef int for comp
 *
 */
typedef int comp2_int;


enum DAY {
    MON = 1, TUE, WED, THU, FRI, SAT, SUN
};
